//BMEmacro.h - Rob Seabrook - May 2020 - for GNU C99 - Target KL27z - IDE MCUXpresso - SDK 2.7.0
//
//Note: r,reg - refers ONLY to a peripheral register - e.g. LPUART0->STAT  (not its address, not a CPU register)
//
//u,uint_type - size and type of reg - can be    uint32_t  uint16_t  uint8_t     as appropriate
//
//    b,lsbit - least significant bit
//      bit0    can be  0 to 31 when u,uint_type is uint32_t
//                      0    15                         16
//                      0     7                          8
//
//    lenbits - number of bits in bit-field 1 to 16    (16 is max even if uint_type is uint32_t)
//   w,width                                ^^^^^^^     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
//
//     m,val  - an unsigned integer value of uint_type or narrower.
//
// SUMMARY of Macros
//    BME_SET(u,r,b)     set bit b  at peripheral register r  of int type u
//    BME_CLR(u,r,b)   clear bit b  at peripheral register r  of int type u
//    BME_TOG(u,r,b)  toggle bit b  at peripheral register r  of int type u  (see also BME_XOR to use mask rather than bit number)
//
//    BME_AND(u,r,m)     read/modify/write peripheral register r  ANDed with mask m  of int type u
//    BME_OR( u,r,m)     read/modify/write peripheral register r  ORed  with mask m  of int type u
//    BME_XOR(u,r,m)     read/modify/write peripheral register r  xored with mask m  of int type u (see also BME_TOG to use bit number)
//    BME_UBFX(u,r,b,w)  assign to or from a bitfield of size w bits starting at bit b in peripheral register r of int type u.
//                       assign from;    value=BME_UBXF(..);
//                       assign to;            BME_UBXF(..)=value;
//    BME_BIT(u,r,b)     special case of BME_UBFX to read 1 bit b(0_31) at peripheral register r of int type u  (result non zero if set, e.g. 001000000b)
//                       example;        value=BME_BIT(uint32_t,CRC0->GPOLY,30); //value is 0x40000000 if bit set

// VERY IMPORTANT - Pitfalls to avoid - confirmed by testing
//
//    1. bit-fields cannot be bigger than 16 bits -- even when integer type is uint32_t !!!!  RTFM
//       (this is because only 4 bits are allowed to hold the bit-field width in the decorated address)
//
//    2. UBFX macro is best used for bit-field reads AND WRITES-also, as follows because it is clearer than using BME_BFI(..)
//           read_data = UBFX(.....);
//                       UBFX(.....) = write_data;   //(because BFI & UBFX address decoration is IDENTICAL)
//
//    3. A bit-field write (ASSIGN TO) expects the sources data bit-field in the same bit position as the destinations
//       example1;   UBFX(uint32_t, CRC0->GPOLY, bit16, width8)=0x003C0000;    //CORRECT if copying the 3C
//       example2;   UBFX(uint32_t, CRC0->GPOLY, bit16, width8)=0x0000003C;    //        This will copy 00   not 3C !!
//       example3;   UBFX(uint8_t,*((uint8_t*)(&CRC0->GPOLY)+2),b0,w8)=0x3C;   //as example1 - this WILL copy 3C too
//
//    4. A bit-field read (ASSIGN FROM) right-justifies (normalizes) the value read
//       example1;   CRC0->GPOLY=0x1200;
//                   value = BME_UBFX(uint32_t,CRC0->GPOLY, 4,8); //value = 0x20 (OK)  (assign from BME_UBFX(..))

#ifndef __BMEmacro_H__
#define __BMEmacro_H__

#include <stddef.h>
#include <stdint.h>

#define BME_LAS1_MASK(BIT)			(3<<26) | (BIT<<21)
#define BME_LAC1_MASK(BIT)			(1<<27) | (BIT<<21)
#define BME_AND_MASK	            (1<<26)
#define BME_OR_MASK                 (1<<27)
#define BME_XOR_MASK                (3<<26)
#define BME_BFI_MASK(BIT,WIDTH)		(1<<28) | (BIT<<23) | ((WIDTH-1)<<19)
#define BME_UBFX_MASK(BIT,WIDTH)	(1<<28) | (BIT<<23) | ((WIDTH-1)<<19)

//         set bit in reg.         example   BME_SET(uint32_t, LPUART0->STAT, LPUART_STAT_OR_SHIFT);
#define    BME_SET( uint_type, reg, bit)     (*(volatile uint_type*)((uint32_t)&(reg) | BME_LAS1_MASK(bit)))
#define    BME_CLR( uint_type, reg, bit)     (*(volatile uint_type*)((uint32_t)&(reg) | BME_LAC1_MASK(bit)))
//see also BME_BIT( uint_type, reg, bit_to_read)
//        AND reg with val.       example   BME_AND(uint32_t, LPUART0->STAT, ~LPUART_STAT_OR_MASK);
#define    BME_AND( uint_type, reg, val)     (*(volatile uint_type*)((uint32_t)&(reg) | BME_AND_MASK) = val)
#define    BME_OR(  uint_type, reg, val)     (*(volatile uint_type*)((uint32_t)&(reg) | BME_OR_MASK)  = val)
#define    BME_XOR( uint_type, reg, val)     (*(volatile uint_type*)((uint32_t)&(reg) | BME_XOR_MASK) = val)
#define    BME_TOG( uint_type, reg, bit)     (*(volatile uint_type*)((uint32_t)&(reg) | BME_XOR_MASK) = (1<<bit))

//Bitfield Insert into reg
//writes val to reg bitfield of len bits starting at lsbit
//                                example   BME_BFI(uint32_t, LPUART0->CTRL, val, len_3, LPUART_CTRL_IDLECFG_SHIFT);
#define    BME_BFI( uint_type, reg, lsbit, width, val) (*(volatile uint_type*)((uint32_t)&(reg) | BME_BFI_MASK(lsbit,width)) = val)
//please note that it is easier to read code that assigns  BME_UBFX(..) = data. Clearer + Fewer args to put in order.

//unsigned bitfield extract from reg.
//returns bitfield in reg of uint_type
//                                example   variable = BME_UBFX(uint32_t, LPUART0->CTRL, LPUART_STAT_OR_SHIFT, 3);
#define    BME_UBFX(uint_type, reg,  lsbit, width) (*(volatile uint_type*)((uint32_t)&(reg) | BME_UBFX_MASK(lsbit,width)))

//read a bit - result is non-zero if bit set e.g. 00000001000000000000000000000000b
#define    BME_BIT( uint_type, reg,  bit_to_read)    BME_UBFX(uint_type, reg, bit_to_read, 1)
#endif
/*
 *   Further notes on BME (Bit Manipulation Engine)
 *
 *   Aim
 *      efficiency in code machine-code size (or/and speed).
 *
 *   General principle
 *       uses the fact that over a limited address range, some of the address bits are always
 *   the same. Those address bits are re-purposed (decorated) to hold additional useful information,
 *   then restored to access the original address - but in the special way that they dictate.
 *
 *   Method
 *      a single assembler "store to memory" or "load from memory" instruction to one of these special
 *   addresses is enough to accomplish the task qualified by the address decoration. The address decoration
 *   is controlled by user calling the appropriate macro with the appropriate parameters.
 *
 *   GPIO
 *      has its own dedicated registers for setting, clearing and toggling bits efficiently.
 *
 *   Safe Testing Suggestion
 *      Use peripheral register CRC0->GPOLY if not already used. The CRC driver would need to be included
 *      in either the new project wizard or the package manager (parcel icon in project explorer),
 *      and the the driver initialisation needs to be called by going to peripheral perspective
 *      and ticking [x]CRC in the Peripherals tab. Remember to reverse these actions after testing
 *      if not using the CRC module in your project.
 *
 *   Disclaimer
 *      this is a work in progress. It has been tested but improvements can always be found. It is
 *      offered free as a goodwill gesture with no warranty of any kind. It is not deemed suitable
 *      for any purpose. You are free to modify and use this file for your own use at your own risk.
 *
 *
 */
